#include "atol.h"

HINSTANCE hInstance;

#ifdef APP_OPT_LOG
int ilog;
#endif

int rnd_seed; // for rnd function.

MAIN_FUNC
{
	hInstance = GetModuleHandle(NULL);

	rnd_seed = (int) (GetTickCount() ^ GetCurrentThreadId());

#ifdef APP_OPT_DEBUG
	console_log("atol (C) 2005, Alexander G. Karpov [mrhx]\n(!) DEBUG BUILD.");
#endif

#ifdef APP_OPT_LOG
	ilog = file_io_open("atol.log", 'w');
#endif

	void load_std_atoms();
	load_std_atoms();

	engine_compile(SCRIPT_FILE);
	engine_run();

#ifdef APP_OPT_DEBUG
	MessageBox(NULL, "Done", "[ atol ]", MB_OK | MB_TOPMOST | MB_ICONINFORMATION);
#endif

#ifdef APP_OPT_LOG
	file_io_close(ilog);
#endif

	ExitProcess(0);

	END_OF_MAIN_FUNC
}

void Error(int IsError, char * Msg)
{
	if (IsError)
	{
		MessageBox(NULL, Msg, "[ atol ] error", MB_OK | MB_TOPMOST | MB_ICONERROR);
		ExitProcess(1);
	}
//#ifdef APP_OPT_DEBUG
//	else console_log("NO_ERR %s", Msg);
//#endif
}

void Error(int IsError, int MsgId, ...)
{
	if (IsError)
	{
		char buf [128], msg [256];

		LoadString(hInstance, MsgId, buf, sizeof(buf));
		wvsprintf(msg, buf, (va_list)(&MsgId+1));

		Error(IsError, msg);
	}
}

#ifndef LINUX_USED

void * operator new(unsigned int n)
{
	register void * p = (void *) LocalAlloc(LMEM_FIXED, n);
	Error(!p, IDS_NO_MEM, n);
	return p;
}

void operator delete(void * p)
{
	if (p) LocalFree(p);
}

#else

void * operator new(unsigned int n)
{
        register void * p = (void *) LocalAlloc(LMEM_FIXED, n);
        Error(!p, IDS_NO_MEM, n);
        return p;
}

void * operator new[](unsigned int n)
{
        register void * p = (void *) LocalAlloc(LMEM_FIXED, n);
        Error(!p, IDS_NO_MEM, n);
        return p;
}

void operator delete(void * p)
{
        if (p) LocalFree(p);
}

void operator delete[](void * p)
{
        if (p) LocalFree(p);
}

long __gxx_personality_v0 = 0;

#endif

int s2i(char *s)
{
	register int r = 0;
	register char *p = s;
	int neg = 0;
	// skip possible space:
	while (*p > 0 && *p <= 32) ++p;

	if (*p == '-') { ++neg; ++p; }
	else if (*p == '+') ++p;

	// check number format:
	if (*p == '0')
	{
		++p;
		if (*p == 'x' || *p == 'X')
		{ // hex:
			for (++p; ; ++p)
				if (*p >= '0' && *p <= '9') r = (r << 4) + (*p - '0');
				else if (*p >= 'a' && *p <= 'f') r = (r << 4) + (*p - ('a' - 10));
				else if (*p >= 'A' && *p <= 'F') r = (r << 4) + (*p - ('A' - 10));
				else break;
		}
		else
		{ // oct:
			for (; *p >= '0' && *p <= '7'; ++p) r = (r << 3) + (*p - '0');
		}
	}
	else
	{ // dec:
		for (; *p >= '0' && *p <= '9'; ++p) r = r*10 + (*p - '0');
	}

	return neg?(-r):r;
}

char * i2s(int x, char * s)
{
	wsprintf(s, "%i", x);
	return s;
}

////// RANDOM: >>>>>>>>>>>>>..

unsigned int rnd()
{
	rnd_seed *= 0x4E35;
	return (unsigned int) ++rnd_seed;
}

//////// LOG ////////////////////

#ifdef APP_OPT_LOG

void Log(char *msg, ...)
{
	char str [MAX_PATH];
	file_io_write(ilog, str, wvsprintf(str, msg, (va_list)(&msg + 1)) );
	file_io_write(ilog, "\r\n", 2);
}

#endif

////////// END OF FILE ////////////////////////
///////////////////////////////////////////////
