#include "atol.h"

// >>> GLOBAL ENGINE VARS: <<<

struct compile_scope_s *ccs = NULL;

int RETURN = 0;		// return flag.    <<***** FOR EXECUTION *****
atom_s *RETURN_R;	// return result.

int is_tracing = 0;	// trace flag.

//// CURRENT SCOPE //////////
scope_s *current_scope = NULL;
scope_s global_scope;

/////////////////////////////

#include "std_func.cpp" // Include 'engine_run_std_func' definition.
#include "std_atoms.cpp" // Include standard atom definition.


// ------ COMPILATION SCOPE CLASS --------

struct compile_scope_s
{
	int current_file; // current file.
	int is_end; // is end of file.
	int is_next_atom; // is there a next atom.
	str_s next_atom; // next atom.
	int next_char;

	// POSITION IN CURRENT FILE:
	int ln_pos;
	char file_name [MAX_PATH];

	compile_scope_s();
	compile_scope_s(compile_scope_s*);
	void done();

protected:
	compile_scope_s *prev;
};

compile_scope_s::compile_scope_s()
{
	current_file	= 0;
	is_end			= 0;
	is_next_atom	= 0;
	next_char		= 0;
	ln_pos			= 1;
	file_name[0]	= 0;
	next_atom.init();
	prev			= ccs;
	ccs				= this;
}

compile_scope_s::compile_scope_s(compile_scope_s *p)
{
	current_file	= p->current_file;
	is_end			= p->is_end;
	is_next_atom	= p->is_next_atom;
	next_char		= p->next_char;
	ln_pos			= p->ln_pos;
	lstrcpy(file_name, p->file_name);
	next_atom.init();
	prev			= ccs;
	ccs				= this;
}

void compile_scope_s::done()
{
	ccs = prev;
	next_atom.free();
}

// ---------------------------------------

/// SYNTAX ERROR FUNCTION ------------------>>>

extern HINSTANCE hInstance;

void S_Error(int IsError, int MsgId, ...)
{
	if (IsError)
	{
		char buf [128], msg [700];

		LoadString(hInstance, MsgId, buf, sizeof(buf)-1);
		wvsprintf(msg, buf, (va_list)(&MsgId+1));
		lstrcat(lstrcat(msg, "\nFile: "), ccs->file_name);
		lstrcat(lstrcat(msg, "\nLine: "), i2s(ccs->ln_pos, buf));

		Error(IsError, msg);
	}
}

///------------------------------<<<<<<


void engine_event(char * event, char * data);

void engine_event(char * event, str_s &data)
{
	engine_event(event, data.str);
}

int engine_read_char()
{
	int chr;
	if (ccs->next_char)
	{
		chr = ccs->next_char;
		ccs->next_char = 0;
	}
	else 
	{
		chr = file_io_get_char(ccs->current_file);
		if (chr == '\n') ++ccs->ln_pos; // check line number.
	}
	if (chr != -1) return chr;
	++ccs->is_end;
	return 0;
}

char * engine_set_next_atom(char * atom)
{
	ccs->next_atom.copy(atom);
	++ccs->is_next_atom;
	return atom;
}

char * engine_set_next_atom(str_s &atom)
{
	ccs->next_atom.copy(&atom);
	++ccs->is_next_atom;
	return atom.str;
}

int is_sep(int ch)
{
	return (ch > 0 && ch <= 32) || ch == ',' || ch == ';';
}

char * engine_read_atom(str_s &atom, int STOPFLAG = 1)
{
	register int chr;

	if (ccs->is_next_atom) // if next atom is set.
	{
		ccs->is_next_atom = 0;
		atom.copy(&ccs->next_atom);
		//console_write("N: <%s>\n", atom);
		//return atom;
		goto END_OF_RA;
	}

	atom.empty();
	chr = engine_read_char();

skip_spaces:

	while (!ccs->is_end && is_sep(chr)) chr = engine_read_char();

	if (ccs->is_end) ;
	else if (chr == '%') // comment support.
	{ // a comment:
		while (!ccs->is_end && chr != '\n' && chr != '\r') chr = engine_read_char();
		goto skip_spaces;
	}
	else if (chr == '#') // multiline comment.
	{
		do chr = engine_read_char(); while (!ccs->is_end && chr != '#'); // do till #
		chr = engine_read_char(); // read next char.
		goto skip_spaces;
	}
	else if (chr == '"' || chr == '\'')
	{ // an atom in quotes:
		int end_chr = chr;

		for (chr = engine_read_char(); !ccs->is_end && chr != end_chr; )
		{
			atom.append((char) chr);
			chr = engine_read_char();
		}
	}
	else if (chr == '(' || chr == ')' || chr == '{' || chr == '}')
	{
		atom.append((char) chr);
	}
	else
	{
		for (; !ccs->is_end && !is_sep(chr) &&
			chr != '(' && chr != ')' && chr != '{' && chr != '}' &&
			chr != '#' && chr != '%' && chr != '"' && chr != '\''; )
		{
			atom.append((char) chr);
			chr = engine_read_char();
		}

		if (!ccs->is_end && !is_sep(chr)) ccs->next_char = chr;
	}

END_OF_RA:

	S_Error(ccs->is_end && *atom.str == 0 && STOPFLAG, IDS_UNEXPECTED_END); // error if unexpected end of file.

	engine_event("atom", atom); // NEW ATOM
	return atom.str;
}

void engine_params()
{
	STR_S(atom);
again:
	engine_read_atom(atom);
	if (atom.neq(')'))
	{
		engine_event("param", atom);
		goto again;
	}
	engine_event("pend", atom); // NEW FUNC PARAMS BEGIN
	/* else ')' - end of params */
	atom.free();
}

void engine_param_list()
{
	STR_S(atom);

	engine_read_atom(atom);
	if (atom.neq('(')) /* not '(' */ engine_set_next_atom(atom);
	else // '('
	{
		engine_event("pbegin", atom); // NEW FUNC PARAMS BEGIN
		// read param name.
		engine_event("param", engine_read_atom(atom)); // NEW FUNC PARAM
		engine_params();
	}
	atom.free();
}

void engine_call();

void engine_call_params()
{
	STR_S(atom);
again:
	engine_read_atom(atom);
	if (atom.neq(')'))
	{
		engine_set_next_atom(atom);
		engine_call();
		goto again;
	}
	/* else ')' - end of params */
	engine_event("cend", atom); // CALL PARAMS END
	atom.free();
}

void engine_call_param_list()
{
	STR_S(atom);

	engine_read_atom(atom);
	if (atom.neq('(')) /* not '(' */ engine_set_next_atom(atom);
	else // '('
	{
		engine_event("cbegin", atom); // CALL PARAMS BEGIN
		engine_call();
		engine_call_params();
	}
	atom.free();
}

void engine_BODY();
void engine_CALLs();

void engine_call()
{
	STR_S(atom);

	engine_event("call", engine_read_atom(atom)); // FUNC CALL NAME

	if (atom.eq("if")) // check for conditional operator.
	{
		engine_event("cbegin", "(");
			engine_call(); // CONDITION
			engine_CALLs(); // THEN
			// ELSE SUPPORT.
			if (lstrcmp(engine_read_atom(atom), "else")) engine_set_next_atom(atom);
			else engine_CALLs();
			
		engine_event("cend", ")");
	}
	else if (atom.eq("with"))
	{
		engine_event("cbegin", "(");
			engine_call();  // PREFIX
			engine_CALLs(); // DO SECTION
		engine_event("cend", ")");
	}
	else engine_call_param_list();
	atom.free();
}

void engine_body()
{
	STR_S(atom);
again:
	engine_set_next_atom( engine_read_atom(atom) );
	if (atom.neq('}'))
	{
		engine_call();
		goto again;
	}
	/* else '}' - end of params */
	atom.free();
}

void engine_BODY()
{
	STR_S(atom);

	engine_read_atom(atom);
	if (atom.neq('{'))
	{
		engine_set_next_atom(atom); // BODY WITH ONLY CALL.

		engine_event("begin", "{");
		engine_call();
		engine_event("end", "}");
	}
	else
	{
		engine_event("begin", atom); // FUNC BODY BEGIN

		engine_body();

		engine_read_atom(atom);
		S_Error(atom.neq('}'), IDS_BODY_END_EXPECTED);
		engine_event("end", atom); // FUNC BODY END
	}
	atom.free();
}

void engine_program();

void engine_CALLs()
{
	STR_S(atom);

	engine_event("call", BODY_STR);
	engine_event("cbegin", "(");

	// check for atom '{' ::
	engine_set_next_atom( engine_read_atom(atom) );
	if (atom.neq('{'))
	{
		engine_call();
		engine_event("cend", ")");
	}
	else engine_BODY();
	atom.free();
}

void engine_func()
{
	STR_S(atom);

	engine_read_atom(atom, 0);

	if (ccs->is_end) return;

	if (atom.eq(std_atom_do->name))
	{
		static int DO_INDEX = 0;
		char di[16];
		atom.append('#')->concat( i2s(++DO_INDEX, di) );

		list_add(&(std_atom_initialize->body), atom.str);

//		atoms_add(&atom)->set_flag(atom_info_system);

		engine_event("atom", atom);
	}

	//console_write("F <%s>\n", atom);

	engine_event("func", atom); // NEW FUNC NAME

	if (atom.eq(std_atom_include->name))
	{
		engine_read_atom(atom);

		S_Error(atom.neq('('), IDS_EXPECTED, "(");
		engine_set_next_atom(atom);

		engine_param_list();
	}
	else if (atom.eq(std_atom_var->name))
	{
		engine_read_atom(atom);

		S_Error(atom.neq('('), IDS_EXPECTED, "(");
		engine_set_next_atom(atom);

		engine_param_list();
	}
	else if (atom.eq(std_atom_new->name))
	{
		engine_read_atom(atom);
		S_Error(atom.neq('('), IDS_EXPECTED, "(");

		engine_event("pbegin", "("); // NEW FUNC PARAMS BEGIN
		// read param name.
		engine_event("param", engine_read_atom(atom)); // VAR NAME
		engine_event("pAram", engine_read_atom(atom)); // VAR VALUE

		engine_read_atom(atom);
		S_Error(atom.neq(')'), IDS_EXPECTED, ")");
	}
	else 
	{
		engine_param_list();
		engine_BODY();
	}
	atom.free();
}

void engine_program()
{
	while (!ccs->is_end) engine_func();
}

void engine_compile(int file, char *fn) ///// COMPILE FILE /////
{
	compile_scope_s c_s;

	c_s.current_file = file;
	lstrcpy(c_s.file_name, fn);

#ifdef APP_OPT_DEBUG
	console_log(">COMPILING...");
#endif

	engine_program();

#ifdef APP_OPT_DEBUG
	console_log(">COMPILED\n");

	atoms_view();
#endif

	after_compile_work();

	c_s.done();
}

void engine_compile(const char *fname) // compile given file
{
	STR_S_(fl, fname);
	fl.concat(".atol");
	int file = file_io_open(fl.str, 'r');

	engine_compile(file, fl.str);

	file_io_close(file);

	fl.free();
}

/**************************************************************************************\

	-- ATOL GRAMMAR --

<program> ::= <function><program> | <.>
<function> ::= <atom><param_list><BODY> | include(<params>) | var(<params>) | do<BODY> |
			   new(<param> <value>)

<param_list> ::= (<atom><params>) | <.>
<params> ::= <atom><params> | <.>
<body> ::= <call><body> | <.>
<BODY> ::= <call> | {<body>}
<call> ::= <atom><call_param_list> | if<call><BODY> | with<call><BODY>
<call_param_list> ::= (<call><call_params>) | <.>
<call_params> ::= <call><call_params> | <.>


\****************************************************************************************/

/////******** PROCESS ENGINE EVENTS ************////

/**
	EVENTS:
	a tom
	f unc
	b egin
	e nd
	pb egin
	pa ram
	pA ram
	pe nd
	ca ll
	cb egin
	ce nd
**/

void engine_event(char * event, char * data)
{
	static atom_s *last_atom = NULL; // last atom.
	static atom_s *current_func = NULL; // current function atom.
	static list_s **current_call[64]; //    .
	static int i_call = -1;
	static list_s *last_call = NULL;
	static list_s ** ptr__;

#ifdef APP_OPT_DEBUG
	console_log("EVENT %s: %s", event, data);
#endif

	switch (*event)
	{
	case 'a':		// atom
		last_atom = atoms_add(data);
		// DEBUG:
		//console_write("ATOM_ADD: <%s>\r\n", data);
		break;

	case 'f':		// func
		S_Error(last_atom->body != NULL &&
			  last_atom != std_atom_var &&
			  last_atom != std_atom_new &&
			  last_atom != std_atom_include, IDS_FUNC_EXISTS, last_atom->name);
		current_func = last_atom;
		current_func->set_flag(atom_info_function); // Set info that this atom is a function.
		break;

	case 'b':		// begin
		S_Error(i_call >= 63, IDS_TOO_MANY_BRACKETS);
		S_Error(current_func == std_atom_include ||
			  current_func == std_atom_var, IDS_UNEXPECTED, "{");

		if (current_func) current_call [++i_call] = &(current_func->body);
		current_func = NULL;
		break;

	case 'e':		// end
		--i_call;
		break;

	case 'p':		// p...
		switch (event [1])
		{
		case 'b':	// pbegin
			break;

		case 'a':	// param
			if (current_func == std_atom_include)
			{   // -- include --
#ifdef APP_OPT_DEBUG
				console_log("INCLUDE %s", data);
#endif
				engine_compile(data);
			}
			else if (current_func == std_atom_var)
			{ // -- add new GLOBAL var --
				//global_scope.new_var(last_atom, std_atom_null);
			  // NEW: add g_var call into atol.initialize::
				list_push( &(std_atom_initialize->body->sublist), last_atom);
			}
			else if (current_func == std_atom_new)
			{
				list_push(ptr__ = &(list_add(&(std_atom_initialize->body), std_atom_g_new)->sublist), atoms_add(data));
			}
			else list_add(&(current_func->args), data);
			break;

		case 'A': // new variable value.
			list_add(ptr__, data);
			break;

		case 'e':	// pend
			break;

		default:
			Error(1, "Unknown event");
		}
		break;

	case 'c':		// c...
		switch (event [1])
		{
		case 'b':	// cbegin
			S_Error(i_call >= 63, IDS_TOO_MANY_BRACKETS);

			current_call [++i_call] = &(last_call->sublist);
			//      
			//     '(',
			//   ,    sublist     .
			break;

		case 'a':	// call
			last_call = list_add(current_call[i_call], data);
			break;

		case 'e':	// cend
			--i_call; //  ,    .
			break;

		default:
			Error(1, "Unknown event");
		}
		break;

	default:
		Error(1, "Unknown event");
	}
}

/*********** ENGINE RUN SECTION ******************/

////// GET FUNC NAME //////////////////////////////////////////////

atom_s * __fastcall get_func_name(list_s * p) //   ,    .
{
	register atom_s *FN; // Real function name.
	char *PREF = current_scope->prefix->name.str; // current prefix.

//	console_write("FN0: %s\n", p->item->name);

	char *p_item_name = p->item->name.str;

//	console_write("p_item_name = %s\n", p_item_name);

	if (*p_item_name == '.') // first dot - ignore prefix.
		FN = atoms_add(p_item_name + 1);
	else if (*PREF) // PREFIX SUPPORT...
	{
		//console_write("PREF\n");
		STR_S_(fn, PREF);
		fn.append('.')->concat(p_item_name);
		if ((FN = atoms_find(fn.str)) && FN->flag(atom_info_function));
		else FN = p->item;
		fn.free();
	}
	else FN = p->item;

//	console_write("FN1: %s\n", FN->name.str);

	return FN;
}

///////////////////////////////////////////////////

list_s * __fastcall engine_run_list(list_s *list) // Run each goal in the list. Return result as a list, You must free it by a call to list_free/1.
{
	list_s *result = NULL, r;

	r.next = NULL;

	for (register list_s *p = list; p; p = p->next)
	{
		r.item = p->item;
		r.sublist = p->sublist;

		list_add(&result, engine_run_func(&r, 1));
	}

	return result;
}

atom_s * __fastcall PROCESS_TRACING(atom_s *func, list_s *args, int&);

atom_s * __fastcall engine_run_func(list_s *p, int FLAG, int RunArgs) // execute function and return result.
// p->item    :: [atom] function name.
// p->sublist :: [list] list of parameters (each item can be a call to another function).
// ======================
// FLAG == 1,    .
// FLAG == 0,      ( ,  then  else).
// ======================
// RunArgs == 1,  .
// RunArgs == 0,   .
// --------------------------------------------------------------------------
{
	if (!p) return std_atom_null;

#ifdef APP_OPT_DEBUG
	console_log("RUN %s/%i %s", p->item->name.str, list_length(p->sublist), current_scope->prefix->name.str);
#endif

	atom_s *FN = get_func_name(p);

	// SPY SUPPORT >>>>>>>>>>>>>
	if (FN->flag(atom_info_spy)) atol_trace(NULL, NULL);

	int is_skipping = 0;

	if (is_tracing && FN != std_atom_body && FN->flag(atom_info_function))
		console_write("%% Calling '%s' ...\n", FN->name.str);

	list_s *args = NULL;
	
	if (RunArgs)
	{
		if (p->item != std_atom_body) args = engine_run_list(p->sublist); // calc all params.
	}
	else args = p->sublist;

#ifdef APP_OPT_DEBUG
	console_log(">ARGS");
	list_view(2, args);
#endif

	{ // TRACING:
		register atom_s* res = PROCESS_TRACING(FN, args, is_skipping);
		if (res) return res;
	}

	atom_s * result = std_atom_null;


#ifdef APP_OPT_DEBUG_RUN
	console_log("%% FN %08X <%s>", FN, FN?FN->name.str:"NULL");
#endif

	list_s *vr = current_scope->find_var(FN);

#ifdef APP_OPT_DEBUG_RUN
	console_log("%% vr %08X", vr);
	if (FN) console_log("%% FN->body %08X ->args %08X", FN->body, FN->args);
	console_log("%% p->sublist %08X", p->sublist);
#endif

	if (vr)
	{
		Error(p->sublist != NULL, IDS_CALL_VAR, FN->name.str);
		result = (atom_s *) vr->sublist;
	}
	else if (FN->notflag(atom_info_function))
	{
		Error(p->sublist != NULL, IDS_CALL_NOT_FUNC, FN->name.str);
		//Error(FN->body != NULL, "internal: not set atom_info_func");
		result = FN; // 'p' is a simple atom.
	}
	else if (FN->flag(atom_info_external))
	{ // it is an external function
#ifdef APP_OPT_DEBUG
		console_log("EXTERN!");
#endif
		result = ( *(atom_s * (WINAPI *)(list_s*, void*))
					FN->body)(args, p->sublist);
	}
	else
	{ // not a variable, it's function

		// step 0: create new scope:
		scope_s scope(&current_scope, std_atom_null, TRUE);

		// check for minimum count of arguments.
		if (list_length(FN->args) > list_length(args))
			Error(1, IDS_AT_LEAST_ARGN, FN->name.str, list_length(FN->args));

		// step 1: set function arguments for call.
		for (list_s *param = FN->args, *arg = args; param;
				param = param->next, arg = arg->next)
		{
			current_scope->new_var(param->item, arg->item);
#ifdef APP_OPT_DEBUG
			console_log("ARG %s=%s", param->item->name.str, arg->item->name.str);
#endif
		}

		// step 2: call the function
		result = engine_run_func(FN->body, 1);

		scope.done(); // step 3: clean memory and pop the scope
	}

	if (RunArgs) list_free(args); // free agrument list.

	if (is_skipping) ++is_tracing;

	if (is_tracing && FN != std_atom_body && FN->flag(atom_info_function))
	{
		console_write("%% Exit: '");
		console_write_str(result->name.str, result->name.len);
		console_write_str("'\n", 2);
	}

	if (RETURN)
	{
		if (FLAG) RETURN = 0;
		result = RETURN_R;
	}
	else
	{
		if (p->next) result = engine_run_func(p->next, FLAG);
	}

	if (!result) result = std_atom_null;

#ifdef APP_OPT_DEBUG
	console_log(">RES %s", result->name);
#endif

	return result;
}

extern atom_s * atom_table;

void engine_run() ///// RUN THE PROGRAM /////
{
#ifdef APP_OPT_DEBUG
	console_log(">RUNNING");
#endif

	if (std_atom_main->notflag(atom_info_function))
	{
#ifdef APP_OPT_DEBUG
	console_log(">NO_MAIN");
#endif

		STR_S_(msg, "WARNING: No 'main' function.\nIf it's not a library then it's error.\n\n----- List of declared functions -----\n\n");
		for (register atom_s * p = atom_table; p; p = p->next)
			if (p->flag(atom_info_function) && p->notflag(atom_info_external) &&
				p != std_atom_initialize &&
				!(p->name.at(0) == 'd' && p->name.at(1) == 'o' && p->name.at(2) == '#')
			   )
				//p->notflag(atom_info_system))
				msg.concat(p->name)->concat(", ");

		MessageBox(NULL, msg.str, "ATOL: Run WARNING", MB_ICONINFORMATION | MB_OK | MB_TOPMOST);
	}
	else
	{
		// DO SOME INIT WORK BEFORE RUNNUNG...
		eval_init();
		//------------------------------------

		current_scope = &global_scope;

		scope_s scope(&current_scope, std_atom_null, TRUE);

		list_s goal;
		goal.next = goal.sublist = NULL;

		/// RUN ATOL.INITIALIZE
		goal.item = std_atom_initialize;
		engine_run_func(&goal, TRUE);

		/// RUN MAIN FUNCTION
		goal.item = std_atom_main;

		engine_run_func(&goal, TRUE);

		scope.done();
	}

#ifdef APP_OPT_DEBUG
	console_log(">DONE");
#endif
}

//// TRACING >>>>>>>>>>>>>>>>>>

atom_s * __fastcall PROCESS_TRACING(atom_s *func, list_s *args, int &is_skipping)
{
	if (!is_tracing || func == std_atom_body || func->notflag(atom_info_function)) return NULL;
	console_write("%% %s: %s( ", (func->flag(atom_info_external)?"Extern":"Call"), func->name.str);

	for (register list_s *p = args; p; p = p->next)
	{
		console_write_str("'", 1);
		console_write_str(p->item->name.str, p->item->name.len);
		console_write_str("' ", 2);
	}

	console_write_str(")\n", 2);

again_chr:
	switch (console_read_char())
	{
	case 'c':
	case 'C':
	case ' ':
		break;

	case 's':
	case 'S':
		++is_skipping;
		is_tracing = 0; // temporary turn off tracing.
		break;

	case 'n':
	case 'N':
		atol_notrace(NULL, NULL);
		break;

	case 'h':
	case 'H':
	case '?':
		console_write("%% Keys: (c)ontinue or space; (s)kip; (n)otrace; (q)uit; (h)elp or ?.\n\n");
		goto again_chr;

	case 'q':
	case 'Q':
		ExitProcess(0);
		break;

	default:
		MessageBeep(0xFFFFFFFF);
		goto again_chr;
	}
	return NULL;
}

