// varied string functions.
#include "str.h"

int str_eqf(const char *a, const char *b)
{
	for (; *a && *b; ++a, ++b)
		if (*a != *b) return 0;
	return 1;
}

char * str_middle(char *dest, char *src, int start, int n) // get N chars of SRC from pos START.
{
	register int i, j;
	for (i = start, j = 0; n > 0 && src[i]; ++i, --n, ++j) dest [j] = src [i];
	dest [j] = 0;
	return dest;
}

char * str_find(char *src, char *substr) // find substring SUBSTR address in SRC.
{
	for (register int i = 0; src [i]; ++i)
		if (!lstrcmp(src + i, substr)) return src + i;
	return NULL;
}

char * str_reverse(char *dest, char *src)
{
	register int i, j;
	for (i = lstrlen(src)-1, j = 0; i >= 0; --i, ++j) dest [j] = src [i];
	dest [j] = 0;
	return dest;
}

int str_like(char *src, char *mask) //    .
{
	register char *p, *m;
	for (p = src, m = mask; *p && *m; ++m, ++p)
		if (*m == '?'); // skip a symbol.
		else if (*m == '*') // any amount of any symbols.
		{
			for (++m; *p; ++p)
				if (str_like(p, m)) return 1;
			break;
		}
		else
		{
			if (*p != *m) break;
		}
	if (*m == '*' && m[1] == 0 && *p == 0) return 1;
	if (*m || *p) return 0;
	return 1;
}

unsigned long str_get_hash(const char *s)
{
	unsigned long hash = 0;
	for (; *s; ++s) hash += *s;
	return hash;
}

/// STR_S STRUCTURE: ------------------------------------------------------------

str_s * str_s::init()
{
	len = size = 0;
	str = new char [1];
	*str = 0;
	hash = 0;
	return this;
}

void str_s::free()
{
	delete [] str;
}

//void str_s::zero()
//{
//	len = size = 0;
//	str = NULL;
//	hash = 0;
//}

str_s * str_s::empty()
{
	len = 0;
	*str = 0;
	hash = 0;
	return this;
}

str_s * str_s::init(const char *s)
{
	init();
	return copy(s);
}

str_s * str_s::init(str_s &s)
{
	init();
	return copy(&s);
}

str_s * str_s::init(int x)
{
	init();
	return convert(x);
}

str_s * str_s::copy(const char *s)
{
	if ((len = lstrlen(s)) > size)
	{
		delete [] str;
		str = new char [(size = len)+1];
	}
	lstrcpy(str, s);
	hash = str_get_hash(s);
	return this;
}

str_s * str_s::concat(const char *s)
{
	len += lstrlen(s);
	if (len > size)
	{
		size = len;
		register char * ptr = new char [size+1];
		lstrcpy(ptr, str);
		delete [] str;
		str = ptr;
	}
	if (str) lstrcat(str, s);
	hash += str_get_hash(s);
	return this;
}

str_s * str_s::append(char ch)
{
	if (++len > size)
	{
		size = len + 16;
		register char * ptr = new char [size+1];
		lstrcpy(ptr, str);
		delete [] str;
		str = ptr;
	}
	str [len-1] = ch;
	str [len] = 0;
	hash += ch;
	return this;

}

str_s * str_s::copy(str_s *s)
{
	if ((len = s->len) > size)
	{
		delete [] str;
		str = new char [(size = len)+1];
	}
	if (str) lstrcpy(str, s->str);
	hash = s->hash;
	return this;
}

str_s * str_s::concat(str_s *s)
{
	len += s->len;
	if (len > size)
	{
		size = len;
		register char * ptr = new char [size+1];
		lstrcpy(ptr, str);
		delete [] str;
		str = ptr;
	}
	lstrcat(str, s->str);
	hash += s->hash;
	return this;
}

str_s * str_s::concat(str_s &s)
{
	return concat(&s);
}

int str_s::eq(const char *s)
{
	return !lstrcmp(str, s);
}

int str_s::neq(const char *s)
{
	return lstrcmp(str, s);
}

int str_s::eqf(const char *s)
{
	return str_eqf(str, s);
}

int str_s::neqf(const char *s)
{
	return !str_eqf(str, s);
}

int str_s::neqi(const char *s)
{
	return lstrcmpi(str, s);
}

int str_s::eq(char ch)
{
	return *str == ch && len == 1;
}

int str_s::neq(char ch)
{
	return *str != ch || len != 1;
}

int str_s::eq(str_s &s)
{
	return hash == s.hash && eq(s.str);
}

int str_s::neq(str_s &s)
{
	return hash != s.hash; // neq(s.str);
}

int str_s::eqf(str_s &s)
{
	return eqf(s.str);
}

int str_s::neqf(str_s &s)
{
	return neqf(s.str);
}

int str_s::neqi(str_s &s)
{
	return neqi(s.str);
}

str_s * str_s::left(int n)
{
	if (len > n) str [len = n] = 0;
	hash = str_get_hash(str);
	return this;
}

str_s * str_s::right(int n)
{
	if (len > n)
	{
		register int i, j;
		for (i = len-n, j = 0; str [i]; ++j, ++i) str [j] = str [i];
		str [j] = 0;
		len = n;
	}
	hash = str_get_hash(str);
	return this;
}

str_s * str_s::move(str_s *s)
{
	delete [] str;
	str  = s->str;
	len  = s->len;
	size = s->size;
	hash = s->hash;
	return this;
}

char str_s::head()
{
	return *str;
}

char * str_s::tail()
{
	return str + 1;
}

char str_s::last()
{
	return at(len - 1);
}

str_s * str_s::convert(int x)
{
	char buf [16];
	wsprintf(buf, "%i", x);
	return copy(buf);
}

extern int s2i(char *s);

int str_s::to_int()
{
	return s2i(str);
}

char str_s::at(int i)
{
	if (i >= len || i < 0) return 0;
	return str [i];
}

str_s * str_s::reverse()
{
	for (register signed int i = 0, j = len-1; i < j; ++i, --j)
	{
		register char t = str [i];
		str [i] = str [j];
		str [j] = t;
	}
	return this;
}

